/* kgx-test-utils.h
 *
 * Copyright 2025 Zander Brown
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <glib.h>

G_BEGIN_DECLS

void            kgx_expect_property_notify               (gpointer    object,
                                                          const char *property);
void            kgx_expect_property_not_notify           (gpointer    object,
                                                          const char *property);
void            kgx_expect_no_notify                     (gpointer    object);
void            kgx_assert_expected_notifies             (gpointer    object);
void            kgx_test_property_notify                 (gpointer    object,
                                                          const char *property,
                                                          ...);

G_END_DECLS
